Documentation.addTranslations({
    "locale": "gl",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "%(filename)s &#8212; %(docstitle)s",
        "&#169; %(copyright_prefix)s %(copyright)s.": "&#169; %(copyright_prefix)s %(copyright)s.",
        ", in ": ", en ",
        "About these documents": "Sobre estes documentos",
        "Automatically generated list of changes in version %(version)s": "Lista xerada automaticamente de cambios na versi\u00f3n %(version)s",
        "C API changes": "Cambios na API C",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "Cambios na versi\u00f3n %(version)s &#8212; %(docstitle)s",
        "Collapse sidebar": "Contraer a barra lateral",
        "Complete Table of Contents": "\u00cdndice completo",
        "Contents": "Contidos",
        "Copyright": "Dereitos de autor\u00eda",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "Creado usando <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.",
        "Expand sidebar": "Expandir a barra lateral",
        "Full index on one page": "\u00cdndice completo nunha p\u00e1xina",
        "General Index": "\u00cdndice xeral",
        "Global Module Index": "\u00cdndice global de m\u00f3dulos",
        "Go": "Ir",
        "Hide Search Matches": "Agochar as coincidencias da busca",
        "Index": "\u00cdndice",
        "Index &#x2013; %(key)s": "",
        "Index pages by letter": "\u00cdndice de p\u00e1xinas por letra",
        "Indices and tables:": "\u00cdndices e t\u00e1boas:",
        "Last updated on %(last_updated)s.": "\u00daltima actualizaci\u00f3n o %(last_updated)s.",
        "Library changes": "Cambios na biblioteca",
        "Navigation": "Navegaci\u00f3n",
        "Next topic": "Seguinte tema",
        "Other changes": "Outros cambios",
        "Overview": "Vista xeral",
        "Please activate JavaScript to enable the search\n    functionality.": "Active JavaScript para activar a funci\u00f3n\n    de busca.",
        "Preparing search...": "Preparando a busca\u2026",
        "Previous topic": "Tema anterior",
        "Quick search": "Busca r\u00e1pida",
        "Search": "Busca",
        "Search Page": "P\u00e1xina de busca",
        "Search Results": "Resultados da busca",
        "Search finished, found one page matching the search query.": [
            "",
            ""
        ],
        "Search within %(docstitle)s": "Buscar dentro do/a %(docstitle)s",
        "Searching": "Buscando",
        "Searching for multiple words only shows matches that contain\n    all words.": "Ao buscar varias palabras s\u00f3 se amosan as coincidencias que\n    conte\u00f1an todas as palabras.",
        "Show Source": "Amosar o c\u00f3digo fonte",
        "Table of Contents": "\u00cdndice",
        "This Page": "Esta p\u00e1xina",
        "Welcome! This is": "D\u00e1moslle a benvida! Isto \u00e9",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "A s\u00faa busca non coincide con ning\u00fan documento. Aseg\u00farese de que todas as palabras estean escritas correctamente e de que seleccionou categor\u00edas abondo.",
        "all functions, classes, terms": "todas as funci\u00f3ns, clases, termos",
        "can be huge": "pode ser enorme",
        "last updated": "\u00faltima actualizaci\u00f3n",
        "lists all sections and subsections": "lista todas as secci\u00f3ns e subsecci\u00f3ns",
        "next chapter": "seguinte cap\u00edtulo",
        "previous chapter": "cap\u00edtulo anterior",
        "quick access to all modules": "acceso r\u00e1pido a todos os m\u00f3dulos",
        "search": "buscar",
        "search this documentation": "buscar esta documentaci\u00f3n",
        "the documentation for": "a documentaci\u00f3n para"
    },
    "plural_expr": "(n != 1)"
});