# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/french/stop.txt

from __future__ import annotations

FRENCH_STOPWORDS = frozenset({
    'ai',
    'aie',
    'aient',
    'aies',
    'ait',
    'as',
    'au',
    'aura',
    'aurai',
    'auraient',
    'aurais',
    'aurait',
    'auras',
    'aurez',
    'auriez',
    'aurions',
    'aurons',
    'auront',
    'aux',
    'avaient',
    'avais',
    'avait',
    'avec',
    'avez',
    'aviez',
    'avions',
    'avons',
    'ayant',
    'ayez',
    'ayons',
    'c',
    'ce',
    'ceci',
    'cela',
    'celà',
    'ces',
    'cet',
    'cette',
    'd',
    'dans',
    'de',
    'des',
    'du',
    'elle',
    'en',
    'es',
    'est',
    'et',
    'eu',
    'eue',
    'eues',
    'eurent',
    'eus',
    'eusse',
    'eussent',
    'eusses',
    'eussiez',
    'eussions',
    'eut',
    'eux',
    'eûmes',
    'eût',
    'eûtes',
    'furent',
    'fus',
    'fusse',
    'fussent',
    'fusses',
    'fussiez',
    'fussions',
    'fut',
    'fûmes',
    'fût',
    'fûtes',
    'ici',
    'il',
    'ils',
    'j',
    'je',
    'l',
    'la',
    'le',
    'les',
    'leur',
    'leurs',
    'lui',
    'm',
    'ma',
    'mais',
    'me',
    'mes',
    'moi',
    'mon',
    'même',
    'n',
    'ne',
    'nos',
    'notre',
    'nous',
    'on',
    'ont',
    'ou',
    'par',
    'pas',
    'pour',
    'qu',
    'que',
    'quel',
    'quelle',
    'quelles',
    'quels',
    'qui',
    's',
    'sa',
    'sans',
    'se',
    'sera',
    'serai',
    'seraient',
    'serais',
    'serait',
    'seras',
    'serez',
    'seriez',
    'serions',
    'serons',
    'seront',
    'ses',
    'soi',
    'soient',
    'sois',
    'soit',
    'sommes',
    'son',
    'sont',
    'soyez',
    'soyons',
    'suis',
    'sur',
    't',
    'ta',
    'te',
    'tes',
    'toi',
    'ton',
    'tu',
    'un',
    'une',
    'vos',
    'votre',
    'vous',
    'y',
    'à',
    'étaient',
    'étais',
    'était',
    'étant',
    'étiez',
    'étions',
    'été',
    'étée',
    'étées',
    'étés',
    'êtes',
})
