Documentation.addTranslations({
    "locale": "en_GB",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "%(filename)s &#8212; %(docstitle)s",
        "&#169; %(copyright_prefix)s %(copyright)s.": "&#169; %(copyright_prefix)s %(copyright)s.",
        ", in ": ", in ",
        "About these documents": "About these documents",
        "Automatically generated list of changes in version %(version)s": "Automatically generated list of changes in version %(version)s",
        "C API changes": "C API changes",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "Changes in Version %(version)s &#8212; %(docstitle)s",
        "Collapse sidebar": "Collapse sidebar",
        "Complete Table of Contents": "Complete Table of Contents",
        "Contents": "Contents",
        "Copyright": "Copyright",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.",
        "Expand sidebar": "Expand sidebar",
        "Full index on one page": "Full index on one page",
        "General Index": "General Index",
        "Global Module Index": "Global Module Index",
        "Go": "Go",
        "Hide Search Matches": "Hide Search Matches",
        "Index": "Index",
        "Index &#x2013; %(key)s": "",
        "Index pages by letter": "Index pages by letter",
        "Indices and tables:": "Indices and tables:",
        "Last updated on %(last_updated)s.": "Last updated on %(last_updated)s.",
        "Library changes": "Library changes",
        "Navigation": "Navigation",
        "Next topic": "Next topic",
        "Other changes": "Other changes",
        "Overview": "Overview",
        "Please activate JavaScript to enable the search\n    functionality.": "Please activate JavaScript to enable the search\n    functionality.",
        "Preparing search...": "Preparing search...",
        "Previous topic": "Previous topic",
        "Quick search": "Quick search",
        "Search": "Search",
        "Search Page": "Search Page",
        "Search Results": "Search Results",
        "Search finished, found one page matching the search query.": [
            "",
            ""
        ],
        "Search within %(docstitle)s": "Search within %(docstitle)s",
        "Searching": "Searching",
        "Searching for multiple words only shows matches that contain\n    all words.": "Searching for multiple words only shows matches that contain\n    all words.",
        "Show Source": "Show Source",
        "Table of Contents": "Table of Contents",
        "This Page": "This Page",
        "Welcome! This is": "Welcome! This is",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.",
        "all functions, classes, terms": "all functions, classes, terms",
        "can be huge": "can be huge",
        "last updated": "last updated",
        "lists all sections and subsections": "lists all sections and subsections",
        "next chapter": "next chapter",
        "previous chapter": "previous chapter",
        "quick access to all modules": "quick access to all modules",
        "search": "search",
        "search this documentation": "search this documentation",
        "the documentation for": "the documentation for"
    },
    "plural_expr": "(n != 1)"
});