Documentation.addTranslations({
    "locale": "cs",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "",
        "&#169; %(copyright_prefix)s %(copyright)s.": "",
        ", in ": ", v ",
        "About these documents": "O t\u011bchto dokumentech",
        "Automatically generated list of changes in version %(version)s": "Automaticky generovan\u00fd seznam zm\u011bn ve verzi %(version)s",
        "C API changes": "Zm\u011bny API",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "",
        "Collapse sidebar": "Sbalit bo\u010dn\u00ed li\u0161tu",
        "Complete Table of Contents": "Celkov\u00fd obsah",
        "Contents": "Obsah",
        "Copyright": "Ve\u0161ker\u00e1 pr\u00e1va vyhrazena",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "",
        "Expand sidebar": "Rozbalit bo\u010dn\u00ed li\u0161tu",
        "Full index on one page": "Cel\u00fd rejst\u0159\u00edk na jedn\u00e9 str\u00e1nce",
        "General Index": "Obecn\u00fd rejst\u0159\u00edk",
        "Global Module Index": "Celkov\u00fd rejst\u0159\u00edk modul\u016f",
        "Go": "OK",
        "Hide Search Matches": "Skr\u00fdt v\u00fdsledky vyhled\u00e1v\u00e1n\u00ed",
        "Index": "Rejst\u0159\u00edk",
        "Index &#x2013; %(key)s": "",
        "Index pages by letter": "Rejst\u0159\u00edk podle p\u00edsmene",
        "Indices and tables:": "Rejst\u0159\u00edky a tabulky:",
        "Last updated on %(last_updated)s.": "Aktualizov\u00e1no dne %(last_updated)s.",
        "Library changes": "Zm\u011bny v knihovn\u00e1ch",
        "Navigation": "Navigace",
        "Next topic": "Dal\u0161\u00ed t\u00e9ma",
        "Other changes": "Ostatn\u00ed zm\u011bny",
        "Overview": "P\u0159ehled",
        "Please activate JavaScript to enable the search\n    functionality.": "Pro podporu vyhled\u00e1v\u00e1n\u00ed aktivujte JavaScript.",
        "Preparing search...": "Vyhled\u00e1v\u00e1n\u00ed se p\u0159ipravuje...",
        "Previous topic": "P\u0159echoz\u00ed t\u00e9ma",
        "Quick search": "Rychl\u00e9 vyhled\u00e1v\u00e1n\u00ed",
        "Search": "Vyhled\u00e1v\u00e1n\u00ed",
        "Search Page": "Vyhled\u00e1vac\u00ed str\u00e1nka",
        "Search Results": "V\u00fdsledky vyhled\u00e1v\u00e1n\u00ed",
        "Search finished, found one page matching the search query.": [
            "",
            "",
            "",
            ""
        ],
        "Search within %(docstitle)s": "Prohledat %(docstitle)s",
        "Searching": "Prob\u00edh\u00e1 vyhled\u00e1n\u00ed",
        "Searching for multiple words only shows matches that contain\n    all words.": "",
        "Show Source": "Uk\u00e1zat zdroj",
        "Table of Contents": "",
        "This Page": "Tato str\u00e1nka",
        "Welcome! This is": "V\u00edtejte! Toto je",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "Vyhled\u00e1v\u00e1n\u00ed nenalezlo \u017e\u00e1dn\u00fd odpov\u00eddaj\u00edc\u00ed dokument. Ujist\u011bte se, \u017ee jste v\u0161echna slova zapsal/a spr\u00e1vn\u011b a \u017ee jste vybral/a dostatek kategori\u00ed.",
        "all functions, classes, terms": "v\u0161echny funkce, t\u0159\u00eddy, term\u00edny",
        "can be huge": "m\u016f\u017ee b\u00fdt obrovsk\u00fd",
        "last updated": "naposledy aktualizov\u00e1no",
        "lists all sections and subsections": "seznam v\u0161ech sekc\u00ed a podsekc\u00ed",
        "next chapter": "dal\u0161\u00ed kapitola",
        "previous chapter": "p\u0159edchoz\u00ed kapitola",
        "quick access to all modules": "rychl\u00fd p\u0159\u00edstup ke v\u0161em modul\u016fm",
        "search": "hledat",
        "search this documentation": "prohledat tuto dokumentaci",
        "the documentation for": "dokumentace pro"
    },
    "plural_expr": "(n == 1 && n % 1 == 0) ? 0 : (n >= 2 && n <= 4 && n % 1 == 0) ? 1: (n % 1 != 0 ) ? 2 : 3"
});