# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/norwegian/stop.txt

from __future__ import annotations

NORWEGIAN_STOPWORDS = frozenset({
    'alle',
    'at',
    'av',
    'bare',
    'begge',
    'ble',
    'blei',
    'bli',
    'blir',
    'blitt',
    'både',
    'båe',
    'da',
    'de',
    'deg',
    'dei',
    'deim',
    'deira',
    'deires',
    'dem',
    'den',
    'denne',
    'der',
    'dere',
    'deres',
    'det',
    'dette',
    'di',
    'din',
    'disse',
    'ditt',
    'du',
    'dykk',
    'dykkar',
    'då',
    'eg',
    'ein',
    'eit',
    'eitt',
    'eller',
    'elles',
    'en',
    'enn',
    'er',
    'et',
    'ett',
    'etter',
    'for',
    'fordi',
    'fra',
    'før',
    'ha',
    'hadde',
    'han',
    'hans',
    'har',
    'hennar',
    'henne',
    'hennes',
    'her',
    'hjå',
    'ho',
    'hoe',
    'honom',
    'hoss',
    'hossen',
    'hun',
    'hva',
    'hvem',
    'hver',
    'hvilke',
    'hvilken',
    'hvis',
    'hvor',
    'hvordan',
    'hvorfor',
    'i',
    'ikke',
    'ikkje',
    'ingen',
    'ingi',
    'inkje',
    'inn',
    'inni',
    'ja',
    'jeg',
    'kan',
    'kom',
    'korleis',
    'korso',
    'kun',
    'kunne',
    'kva',
    'kvar',
    'kvarhelst',
    'kven',
    'kvi',
    'kvifor',
    'man',
    'mange',
    'me',
    'med',
    'medan',
    'meg',
    'meget',
    'mellom',
    'men',
    'mi',
    'min',
    'mine',
    'mitt',
    'mot',
    'mykje',
    'ned',
    'no',
    'noe',
    'noen',
    'noka',
    'noko',
    'nokon',
    'nokor',
    'nokre',
    'nå',
    'når',
    'og',
    'også',
    'om',
    'opp',
    'oss',
    'over',
    'på',
    'samme',
    'seg',
    'selv',
    'si',
    'sia',
    'sidan',
    'siden',
    'sin',
    'sine',
    'sitt',
    'sjøl',
    'skal',
    'skulle',
    'slik',
    'so',
    'som',
    'somme',
    'somt',
    'så',
    'sånn',
    'til',
    'um',
    'upp',
    'ut',
    'uten',
    'var',
    'vart',
    'varte',
    'ved',
    'vere',
    'verte',
    'vi',
    'vil',
    'ville',
    'vore',
    'vors',
    'vort',
    'vår',
    'være',
    'vært',
    'å',
})
