Documentation.addTranslations({
    "locale": "fi",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "",
        "&#169; %(copyright_prefix)s %(copyright)s.": "",
        ", in ": "",
        "About these documents": "Tietoja t\u00e4st\u00e4 documentist\u00e4",
        "Automatically generated list of changes in version %(version)s": "Automaattisesti luotu muutoshistoria alkaen versiosta %(version)s",
        "C API changes": "",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "",
        "Collapse sidebar": "",
        "Complete Table of Contents": "",
        "Contents": "",
        "Copyright": "",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "",
        "Expand sidebar": "",
        "Full index on one page": "Hakemisto yhten\u00e4 luettelona",
        "General Index": "Yleinen sis\u00e4llysluettelo",
        "Global Module Index": "Yleinen moduulien sis\u00e4llysluettelo",
        "Go": "Siirry",
        "Hide Search Matches": "Piilota l\u00f6ydetyt",
        "Index": "Sis\u00e4llysluettelo",
        "Index &#x2013; %(key)s": "",
        "Index pages by letter": "Hakemisto aakkostus sivuttain",
        "Indices and tables:": "",
        "Last updated on %(last_updated)s.": "",
        "Library changes": "",
        "Navigation": "Navikointi",
        "Next topic": ">>",
        "Other changes": "",
        "Overview": "Yhteenveto",
        "Please activate JavaScript to enable the search\n    functionality.": "Javascript pit\u00e4\u00e4 olla sallittu, jotta etsint\u00e4 toimii.",
        "Preparing search...": "",
        "Previous topic": "<<",
        "Quick search": "Pikahaku",
        "Search": "Etsi",
        "Search Page": "Etsi sivu",
        "Search Results": "Etsinn\u00e4n tulos",
        "Search finished, found one page matching the search query.": [
            "",
            ""
        ],
        "Search within %(docstitle)s": "",
        "Searching": "",
        "Searching for multiple words only shows matches that contain\n    all words.": "",
        "Show Source": "N\u00e4yt\u00e4 l\u00e4hdekoodina",
        "Table of Contents": "",
        "This Page": "T\u00e4m\u00e4 sivu",
        "Welcome! This is": "",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "",
        "all functions, classes, terms": "",
        "can be huge": "voi olla iso",
        "last updated": "",
        "lists all sections and subsections": "",
        "next chapter": ">>",
        "previous chapter": "<<",
        "quick access to all modules": "",
        "search": "etsi",
        "search this documentation": "",
        "the documentation for": ""
    },
    "plural_expr": "(n != 1)"
});