Documentation.addTranslations({
    "locale": "sq",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "%(filename)s &#8212; %(docstitle)s",
        "&#169; %(copyright_prefix)s %(copyright)s.": "&#169; %(copyright_prefix)s %(copyright)s.",
        ", in ": ", n\u00eb ",
        "About these documents": "Mbi k\u00ebto dokumente",
        "Automatically generated list of changes in version %(version)s": "List\u00eb ndryshime n\u00eb versionin %(version)s e prodhuar automatikisht",
        "C API changes": "Ndryshime API C",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "Ndryshe n\u00eb Versionin %(version)s &#8212; %(docstitle)s",
        "Collapse sidebar": "Tkurre an\u00ebshtyll\u00ebn",
        "Complete Table of Contents": "Tryez\u00eb e Plot\u00eb e L\u00ebnd\u00ebs",
        "Contents": "L\u00ebnd\u00eb",
        "Copyright": "T\u00eb drejta kopjimi",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "Krijuar duke p\u00ebrdorur <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.",
        "Expand sidebar": "Zgjeroje an\u00ebshtyll\u00ebn",
        "Full index on one page": "Tregues i plot\u00eb n\u00eb nj\u00eb faqe",
        "General Index": "Tregues i P\u00ebrgjithsh\u00ebm",
        "Global Module Index": "Tregues Global Modulesh",
        "Go": "Shko",
        "Hide Search Matches": "Fshih P\u00ebrputhje K\u00ebrkimi",
        "Index": "Tregues",
        "Index &#x2013; %(key)s": "Tregues &#x2013; %(key)s",
        "Index pages by letter": "Faqe treguesi sipas shkronjash",
        "Indices and tables:": "Tregues dhe tabela:",
        "Last updated on %(last_updated)s.": "P\u00ebrdit\u00ebsuar s\u00eb fundi m\u00eb %(last_updated)s.",
        "Library changes": "Ndryshime librarie",
        "Navigation": "L\u00ebvizje",
        "Next topic": "Subjekti pasues",
        "Other changes": "Ndryshime t\u00eb tjera",
        "Overview": "P\u00ebrmbledhje",
        "Please activate JavaScript to enable the search\n    functionality.": "Q\u00eb t\u00eb aktivizohet funksioni i k\u00ebrkimit, ju lutemi, aktivizoni\n    JavaScript-in.",
        "Preparing search...": "Po p\u00ebrgatitet k\u00ebrkim...",
        "Previous topic": "Subjekti i m\u00ebparsh\u00ebm",
        "Quick search": "K\u00ebrkim i shpejt\u00eb",
        "Search": "K\u0113rko",
        "Search Page": "Faqe K\u00ebrkimesh",
        "Search Results": "P\u00ebrfundime K\u00ebrkimi",
        "Search finished, found one page matching the search query.": [
            "K\u00ebrkimi p\u00ebrfundoi, u gjet nj\u00eb faqe me p\u00ebrkime p\u00ebr termat e k\u00ebrkimit.",
            "K\u00ebrkimi p\u00ebrfundoi, u gjet\u00ebn ${resultCount} faqe me p\u00ebrkime p\u00ebr termat e k\u00ebrkimit."
        ],
        "Search within %(docstitle)s": "K\u00ebrkoni brenda %(docstitle)s",
        "Searching": "K\u00ebrkim",
        "Searching for multiple words only shows matches that contain\n    all words.": "K\u00ebrkimi p\u00ebr disa fjal\u00eb nj\u00ebher\u00ebsh shfaq vet\u00ebm p\u00ebrputhje q\u00eb\n    p\u00ebrmbajn\u00eb krejt fjal\u00ebt.",
        "Show Source": "Shfaq Burimin",
        "Table of Contents": "Tryeza e L\u00ebnd\u00ebs",
        "This Page": "Kjo Faqe",
        "Welcome! This is": "Mir\u00eb se vini! Ky \u00ebsht\u00eb",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "K\u00ebrkimi juaj s\u2019gjeti p\u00ebrputhje me ndonj\u00eb dokument. Ju lutemi, sigurohuni se jan\u00eb shkruar sakt\u00eb krejt fjal\u00ebt dhe se keni p\u00ebrzgjedhur aq kategori sa duhen.",
        "all functions, classes, terms": "krejt funksionet, klasat, termat",
        "can be huge": "mund t\u00eb jet\u00eb i st\u00ebrmadh",
        "last updated": "p\u00ebrdit\u00ebsuar s\u00eb fundi m\u00eb",
        "lists all sections and subsections": "paraqet krejt ndarjet dhe n\u00ebnndarjet",
        "next chapter": "kapitulli pasues",
        "previous chapter": "kapitulli i m\u00ebparsh\u00ebm",
        "quick access to all modules": "hyrje e shpejt\u00eb te krejt modulet",
        "search": "k\u00ebrko",
        "search this documentation": "k\u00ebrkoni te ky dokumentim",
        "the documentation for": "dokumentimi p\u00ebr"
    },
    "plural_expr": "(n != 1)"
});