var DanishStemmer=function(){var s=new BaseStemmer,c=[["hed",-1,1],["ethed",0,1],["ered",-1,1],["e",-1,1],["erede",3,1],["ende",3,1],["erende",5,1],["ene",3,1],["erne",3,1],["ere",3,1],["en",-1,1],["heden",10,1],["eren",10,1],["er",-1,1],["heder",13,1],["erer",13,1],["s",-1,2],["heds",16,1],["es",16,1],["endes",18,1],["erendes",19,1],["enes",18,1],["ernes",18,1],["eres",18,1],["ens",16,1],["hedens",24,1],["erens",24,1],["ers",16,1],["ets",16,1],["erets",28,1],["et",-1,1],["eret",30,1]],e=[["gd",-1,-1],["dt",-1,-1],["gt",-1,-1],["kt",-1,-1]],t=[["ig",-1,1],["lig",0,1],["elig",1,1],["els",-1,1],["løst",-1,2]],a=[119,223,119,1],o=[17,65,16,1,0,0,0,0,0,0,0,0,0,0,0,0,48,0,128],u=[239,254,42,3,0,0,0,0,0,0,0,0,0,0,0,0,16],l=0,m=0,_="";function n(){var r,i=s.limit-s.cursor;s.cursor<m||(r=s.limit_backward,s.limit_backward=m,s.ket=s.cursor,0==s.find_among_b(e)?s.limit_backward=r:(s.bra=s.cursor,s.limit_backward=r,s.cursor=s.limit-i,s.cursor<=s.limit_backward||(s.cursor--,s.bra=s.cursor,s.slice_del())))}this.stem=function(){var r=s.cursor,i=(m=s.limit,e=s.cursor,(i=s.cursor+3)>s.limit||(s.cursor=i,l=s.cursor,s.cursor=e,s.go_out_grouping(o,97,248)&&(s.cursor++,s.go_in_grouping(o,97,248))&&(s.cursor++,m=s.cursor,l<=m||(m=l))),s.cursor=r,s.limit_backward=s.cursor,s.cursor=s.limit,s.limit-s.cursor),e=((()=>{var r;if(!(s.cursor<m)){var i=s.limit_backward;if(s.limit_backward=m,s.ket=s.cursor,0==(r=s.find_among_b(c)))s.limit_backward=i;else switch(s.bra=s.cursor,s.limit_backward=i,r){case 1:if(s.slice_del())break;return;case 2:if(!s.in_grouping_b(u,97,229))return;if(s.slice_del())break}}})(),s.cursor=s.limit-i,s.limit-s.cursor),r=(n(),s.cursor=s.limit-e,s.limit-s.cursor),i=((()=>{var r,i=s.limit-s.cursor;if(s.ket=s.cursor,(!s.eq_s_b("st")||(s.bra=s.cursor,!s.eq_s_b("ig"))||s.slice_del())&&(s.cursor=s.limit-i,!(s.cursor<m))){i=s.limit_backward;if(s.limit_backward=m,s.ket=s.cursor,0==(r=s.find_among_b(t)))s.limit_backward=i;else switch(s.bra=s.cursor,s.limit_backward=i,r){case 1:if(!s.slice_del())return;var e=s.limit-s.cursor;n(),s.cursor=s.limit-e;break;case 2:if(s.slice_from("løs"))break}}})(),s.cursor=s.limit-r,s.limit-s.cursor);return s.cursor<m||(e=s.limit_backward,s.limit_backward=m,s.ket=s.cursor,s.in_grouping_b(a,98,122)?(s.bra=s.cursor,""!=(_=s.slice_to())&&(s.limit_backward=e,s.eq_s_b(_))&&s.slice_del()):s.limit_backward=e),s.cursor=s.limit-i,s.cursor=s.limit_backward,!0},this.stemWord=function(r){return s.setCurrent(r),this.stem(),s.getCurrent()}};