# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/swedish/stop.txt

from __future__ import annotations

SWEDISH_STOPWORDS = frozenset({
    'alla',
    'allt',
    'att',
    'av',
    'blev',
    'bli',
    'blir',
    'blivit',
    'de',
    'dem',
    'den',
    'denna',
    'deras',
    'dess',
    'dessa',
    'det',
    'detta',
    'dig',
    'din',
    'dina',
    'ditt',
    'du',
    'där',
    'då',
    'efter',
    'ej',
    'eller',
    'en',
    'er',
    'era',
    'ert',
    'ett',
    'från',
    'för',
    'ha',
    'hade',
    'han',
    'hans',
    'har',
    'henne',
    'hennes',
    'hon',
    'honom',
    'hur',
    'här',
    'i',
    'icke',
    'ingen',
    'inom',
    'inte',
    'jag',
    'ju',
    'kan',
    'kunde',
    'man',
    'med',
    'mellan',
    'men',
    'mig',
    'min',
    'mina',
    'mitt',
    'mot',
    'mycket',
    'ni',
    'nu',
    'när',
    'någon',
    'något',
    'några',
    'och',
    'om',
    'oss',
    'på',
    'samma',
    'sedan',
    'sig',
    'sin',
    'sina',
    'sitta',
    'själv',
    'skulle',
    'som',
    'så',
    'sådan',
    'sådana',
    'sådant',
    'till',
    'under',
    'upp',
    'ut',
    'utan',
    'vad',
    'var',
    'vara',
    'varför',
    'varit',
    'varje',
    'vars',
    'vart',
    'vem',
    'vi',
    'vid',
    'vilka',
    'vilkas',
    'vilken',
    'vilket',
    'vår',
    'våra',
    'vårt',
    'än',
    'är',
    'åt',
    'över',
})
