# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/german/stop.txt

from __future__ import annotations

GERMAN_STOPWORDS = frozenset({
    'aber',
    'alle',
    'allem',
    'allen',
    'aller',
    'alles',
    'als',
    'also',
    'am',
    'an',
    'ander',
    'andere',
    'anderem',
    'anderen',
    'anderer',
    'anderes',
    'anderm',
    'andern',
    'anderr',
    'anders',
    'auch',
    'auf',
    'aus',
    'bei',
    'bin',
    'bis',
    'bist',
    'da',
    'damit',
    'dann',
    'das',
    'daß',
    'dasselbe',
    'dazu',
    'dein',
    'deine',
    'deinem',
    'deinen',
    'deiner',
    'deines',
    'dem',
    'demselben',
    'den',
    'denn',
    'denselben',
    'der',
    'derer',
    'derselbe',
    'derselben',
    'des',
    'desselben',
    'dessen',
    'dich',
    'die',
    'dies',
    'diese',
    'dieselbe',
    'dieselben',
    'diesem',
    'diesen',
    'dieser',
    'dieses',
    'dir',
    'doch',
    'dort',
    'du',
    'durch',
    'ein',
    'eine',
    'einem',
    'einen',
    'einer',
    'eines',
    'einig',
    'einige',
    'einigem',
    'einigen',
    'einiger',
    'einiges',
    'einmal',
    'er',
    'es',
    'etwas',
    'euch',
    'euer',
    'eure',
    'eurem',
    'euren',
    'eurer',
    'eures',
    'für',
    'gegen',
    'gewesen',
    'hab',
    'habe',
    'haben',
    'hat',
    'hatte',
    'hatten',
    'hier',
    'hin',
    'hinter',
    'ich',
    'ihm',
    'ihn',
    'ihnen',
    'ihr',
    'ihre',
    'ihrem',
    'ihren',
    'ihrer',
    'ihres',
    'im',
    'in',
    'indem',
    'ins',
    'ist',
    'jede',
    'jedem',
    'jeden',
    'jeder',
    'jedes',
    'jene',
    'jenem',
    'jenen',
    'jener',
    'jenes',
    'jetzt',
    'kann',
    'kein',
    'keine',
    'keinem',
    'keinen',
    'keiner',
    'keines',
    'können',
    'könnte',
    'machen',
    'man',
    'manche',
    'manchem',
    'manchen',
    'mancher',
    'manches',
    'mein',
    'meine',
    'meinem',
    'meinen',
    'meiner',
    'meines',
    'mich',
    'mir',
    'mit',
    'muss',
    'musste',
    'nach',
    'nicht',
    'nichts',
    'noch',
    'nun',
    'nur',
    'ob',
    'oder',
    'ohne',
    'sehr',
    'sein',
    'seine',
    'seinem',
    'seinen',
    'seiner',
    'seines',
    'selbst',
    'sich',
    'sie',
    'sind',
    'so',
    'solche',
    'solchem',
    'solchen',
    'solcher',
    'solches',
    'soll',
    'sollte',
    'sondern',
    'sonst',
    'um',
    'und',
    'uns',
    'unse',
    'unsem',
    'unsen',
    'unser',
    'unses',
    'unter',
    'viel',
    'vom',
    'von',
    'vor',
    'war',
    'waren',
    'warst',
    'was',
    'weg',
    'weil',
    'weiter',
    'welche',
    'welchem',
    'welchen',
    'welcher',
    'welches',
    'wenn',
    'werde',
    'werden',
    'wie',
    'wieder',
    'will',
    'wir',
    'wird',
    'wirst',
    'wo',
    'wollen',
    'wollte',
    'während',
    'würde',
    'würden',
    'zu',
    'zum',
    'zur',
    'zwar',
    'zwischen',
    'über',
})
