# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/russian/stop.txt

from __future__ import annotations

RUSSIAN_STOPWORDS = frozenset({
    'а',
    'без',
    'более',
    'больше',
    'будет',
    'будто',
    'бы',
    'был',
    'была',
    'были',
    'было',
    'быть',
    'в',
    'вам',
    'вас',
    'вдруг',
    'ведь',
    'во',
    'вот',
    'впрочем',
    'все',
    'всегда',
    'всего',
    'всех',
    'всю',
    'вы',
    'где',
    'говорил',
    'да',
    'даже',
    'два',
    'для',
    'до',
    'другой',
    'его',
    'ее',
    'ей',
    'ему',
    'если',
    'есть',
    'еще',
    'ж',
    'же',
    'жизнь',
    'за',
    'зачем',
    'здесь',
    'и',
    'из',
    'или',
    'им',
    'иногда',
    'их',
    'к',
    'кажется',
    'как',
    'какая',
    'какой',
    'когда',
    'конечно',
    'кто',
    'куда',
    'ли',
    'лучше',
    'между',
    'меня',
    'мне',
    'много',
    'может',
    'можно',
    'мой',
    'моя',
    'мы',
    'на',
    'над',
    'надо',
    'наконец',
    'нас',
    'не',
    'него',
    'нее',
    'ней',
    'нельзя',
    'нет',
    'ни',
    'нибудь',
    'никогда',
    'ним',
    'них',
    'ничего',
    'но',
    'ну',
    'о',
    'об',
    'один',
    'он',
    'она',
    'они',
    'опять',
    'от',
    'перед',
    'по',
    'под',
    'после',
    'потом',
    'потому',
    'почти',
    'при',
    'про',
    'раз',
    'разве',
    'с',
    'сам',
    'свою',
    'себе',
    'себя',
    'сегодня',
    'сейчас',
    'сказал',
    'сказала',
    'сказать',
    'со',
    'совсем',
    'так',
    'такой',
    'там',
    'тебя',
    'тем',
    'теперь',
    'то',
    'тогда',
    'того',
    'тоже',
    'только',
    'том',
    'тот',
    'три',
    'тут',
    'ты',
    'у',
    'уж',
    'уже',
    'хорошо',
    'хоть',
    'чего',
    'человек',
    'чем',
    'через',
    'что',
    'чтоб',
    'чтобы',
    'чуть',
    'эти',
    'этого',
    'этой',
    'этом',
    'этот',
    'эту',
    'я',
})
