Documentation.addTranslations({
    "locale": "sv",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "%(filename)s - %(docstitle)s",
        "&#169; %(copyright_prefix)s %(copyright)s.": "\u00a9 %(copyright_prefix)s %(copyright)s.",
        ", in ": ", i ",
        "About these documents": "Om dessa dokument",
        "Automatically generated list of changes in version %(version)s": "Automatiskt genererad lista \u00f6ver \u00e4ndringar i version %(version)s",
        "C API changes": "C API-\u00e4ndringar",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "\u00c4ndringar i version %(version)s - %(docstitle)s",
        "Collapse sidebar": "D\u00f6lja sidof\u00e4ltet",
        "Complete Table of Contents": "Fullst\u00e4ndig inneh\u00e5llsf\u00f6rteckning",
        "Contents": "Inneh\u00e5ll",
        "Copyright": "Upphovsr\u00e4tt",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "Skapad med hj\u00e4lp av <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.",
        "Expand sidebar": "Visa sidof\u00e4ltet",
        "Full index on one page": "Fullst\u00e4ndigt index p\u00e5 en sida",
        "General Index": "Allm\u00e4nt index",
        "Global Module Index": "Globalt modulindex",
        "Go": "Start",
        "Hide Search Matches": "D\u00f6lj s\u00f6ktr\u00e4ffar",
        "Index": "Index",
        "Index &#x2013; %(key)s": "Index - %(key)s",
        "Index pages by letter": "Indexsidor per bokstav",
        "Indices and tables:": "Index och tabeller:",
        "Last updated on %(last_updated)s.": "Senast uppdaterad den %(last_updated)s.",
        "Library changes": "F\u00f6r\u00e4ndringar i biblioteket",
        "Navigation": "Navigering",
        "Next topic": "N\u00e4sta \u00e4mne",
        "Other changes": "\u00d6vriga \u00e4ndringar",
        "Overview": "\u00d6versikt",
        "Please activate JavaScript to enable the search\n    functionality.": "Aktivera JavaScript f\u00f6r att aktivera s\u00f6k-\n    funktionalitet.",
        "Preparing search...": "F\u00f6rbereder s\u00f6kning...",
        "Previous topic": "F\u00f6reg\u00e5ende \u00e4mne",
        "Quick search": "Snabbs\u00f6k",
        "Search": "S\u00f6k",
        "Search Page": "S\u00f6ksida",
        "Search Results": "S\u00f6kresultat",
        "Search finished, found one page matching the search query.": [
            "S\u00f6kningen \u00e4r klar, hittade en sida som matchar s\u00f6kfr\u00e5gan.",
            "S\u00f6kningen \u00e4r klar, hittade ${resultCount} sidor som matchar s\u00f6kfr\u00e5gan."
        ],
        "Search within %(docstitle)s": "S\u00f6k inom %(docstitle)s",
        "Searching": "S\u00f6ker",
        "Searching for multiple words only shows matches that contain\n    all words.": "Vid s\u00f6kning efter flera ord visas endast matchningar som inneh\u00e5ller\n    alla ord.",
        "Show Source": "Visa k\u00e4lla",
        "Table of Contents": "Inneh\u00e5ll",
        "This Page": "Den h\u00e4r sidan",
        "Welcome! This is": "V\u00e4lkomna! Det h\u00e4r \u00e4r",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "Din s\u00f6kning matchade inte n\u00e5gra dokument. Kontrollera att alla ord \u00e4r r\u00e4ttstavade och att du har valt tillr\u00e4ckligt m\u00e5nga kategorier.",
        "all functions, classes, terms": "alla funktioner, klasser, termer",
        "can be huge": "kan vara enorma",
        "last updated": "senast uppdaterad",
        "lists all sections and subsections": "listar alla avsnitt och underavsnitt",
        "next chapter": "n\u00e4sta kapitel",
        "previous chapter": "f\u00f6reg\u00e5ende kapitel",
        "quick access to all modules": "snabb \u00e5tkomst till alla moduler",
        "search": "s\u00f6k",
        "search this documentation": "s\u00f6k i denna dokumentation",
        "the documentation for": "dokumentationen f\u00f6r"
    },
    "plural_expr": "(n != 1)"
});