Documentation.addTranslations({
    "locale": "de",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "",
        "&#169; %(copyright_prefix)s %(copyright)s.": "",
        ", in ": ", in ",
        "About these documents": "\u00dcber dieses Dokument",
        "Automatically generated list of changes in version %(version)s": "Automatisch generierte Liste der \u00c4nderungen in Version %(version)s",
        "C API changes": "C API-\u00c4nderungen",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "",
        "Collapse sidebar": "Seitenleiste einklappen",
        "Complete Table of Contents": "Vollst\u00e4ndiges Inhaltsverzeichnis",
        "Contents": "Inhalt",
        "Copyright": "Copyright",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "",
        "Expand sidebar": "Seitenleiste ausklappen",
        "Full index on one page": "Gesamtes Stichwortverzeichnis auf einer Seite",
        "General Index": "Stichwortverzeichnis",
        "Global Module Index": "Globaler Modulindex",
        "Go": "Los",
        "Hide Search Matches": "Suchergebnisse ausblenden",
        "Index": "Stichwortverzeichnis",
        "Index &#x2013; %(key)s": "",
        "Index pages by letter": "Stichwortverzeichnis nach Anfangsbuchstabe",
        "Indices and tables:": "Verzeichnisse und Tabellen:",
        "Last updated on %(last_updated)s.": "Zuletzt aktualisiert am %(last_updated)s.",
        "Library changes": "Bibliotheks-\u00c4nderungen",
        "Navigation": "Navigation",
        "Next topic": "N\u00e4chstes Thema",
        "Other changes": "Andere \u00c4nderungen",
        "Overview": "\u00dcbersicht",
        "Please activate JavaScript to enable the search\n    functionality.": "Bitte aktivieren Sie JavaScript, wenn Sie die Suchfunktion nutzen wollen.",
        "Preparing search...": "Suche wird vorbereitet...",
        "Previous topic": "Vorheriges Thema",
        "Quick search": "Schnellsuche",
        "Search": "Suche",
        "Search Page": "Suche",
        "Search Results": "Suchergebnisse",
        "Search finished, found one page matching the search query.": [
            "",
            ""
        ],
        "Search within %(docstitle)s": "Suche in %(docstitle)s",
        "Searching": "Suchen",
        "Searching for multiple words only shows matches that contain\n    all words.": "",
        "Show Source": "Quellcode anzeigen",
        "Table of Contents": "Inhaltsverzeichnis",
        "This Page": "Diese Seite",
        "Welcome! This is": "Willkommen! Dies ist",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "Ihre Suche ergab keine Treffer. Bitte stellen Sie sicher, dass alle W\u00f6rter richtig geschrieben sind und gen\u00fcgend Kategorien ausgew\u00e4hlt sind.",
        "all functions, classes, terms": "alle Funktionen, Klassen, Begriffe",
        "can be huge": "kann gro\u00df sein",
        "last updated": "zuletzt aktualisiert",
        "lists all sections and subsections": "Liste aller Kapitel und Unterkapitel",
        "next chapter": "n\u00e4chstes Kapitel",
        "previous chapter": "vorheriges Kapitel",
        "quick access to all modules": "schneller Zugriff auf alle Module",
        "search": "suchen",
        "search this documentation": "durchsuche diese Dokumentation",
        "the documentation for": "die Dokumentation f\u00fcr"
    },
    "plural_expr": "(n != 1)"
});