Documentation.addTranslations({
    "locale": "zh_Hant_TW",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "%(filename)s &#8212; %(docstitle)s",
        "&#169; %(copyright_prefix)s %(copyright)s.": "&#169; %(copyright_prefix)s %(copyright)s.",
        ", in ": "\uff0c\u65bc ",
        "About these documents": "\u95dc\u65bc\u9019\u4e9b\u6587\u4ef6",
        "Automatically generated list of changes in version %(version)s": "\u81ea\u52d5\u7522\u751f\u7684 %(version)s \u7248\u8b8a\u66f4\u5217\u8868",
        "C API changes": "C API \u7684\u8b8a\u66f4",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "\u65bc %(version)s \u7248\u4e2d\u7684\u6240\u6709\u8b8a\u66f4 &#8212; %(docstitle)s",
        "Collapse sidebar": "\u6536\u5408\u5074\u908a\u6b04",
        "Complete Table of Contents": "\u5b8c\u6574\u76ee\u9304",
        "Contents": "\u5167\u5bb9",
        "Copyright": "\u7248\u6b0a\u6240\u6709",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "\u4f7f\u7528 <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s \u5efa\u7acb\u3002",
        "Expand sidebar": "\u5c55\u958b\u5074\u908a\u6b04",
        "Full index on one page": "\u55ae\u9801\u5b8c\u6574\u7d22\u5f15",
        "General Index": "\u7e3d\u7d22\u5f15",
        "Global Module Index": "\u5168\u57df\u6a21\u7d44\u7d22\u5f15",
        "Go": "\u524d\u5f80",
        "Hide Search Matches": "\u96b1\u85cf\u7b26\u5408\u641c\u5c0b",
        "Index": "\u7d22\u5f15",
        "Index &#x2013; %(key)s": "",
        "Index pages by letter": "\u6309\u5b57\u6bcd\u7d22\u5f15\u9801\u9762",
        "Indices and tables:": "\u7d22\u5f15\u8207\u8868\u683c\uff1a",
        "Last updated on %(last_updated)s.": "\u6700\u5f8c\u66f4\u65b0\u65bc %(last_updated)s\u3002",
        "Library changes": "\u7a0b\u5f0f\u5eab\u7684\u8b8a\u66f4",
        "Navigation": "\u700f\u89bd",
        "Next topic": "\u4e0b\u500b\u4e3b\u984c",
        "Other changes": "\u5176\u4ed6\u8b8a\u66f4",
        "Overview": "\u6982\u8981",
        "Please activate JavaScript to enable the search\n    functionality.": "\u8acb\u555f\u7528 Javascript \u4ee5\u958b\u555f\u641c\u5c0b\u529f\u80fd\u3002",
        "Preparing search...": "\u6e96\u5099\u641c\u5c0b\u4e2d\u2026",
        "Previous topic": "\u4e0a\u500b\u4e3b\u984c",
        "Quick search": "\u5feb\u901f\u641c\u5c0b",
        "Search": "\u641c\u5c0b",
        "Search Page": "\u641c\u5c0b\u9801\u9762",
        "Search Results": "\u641c\u5c0b\u7d50\u679c",
        "Search finished, found one page matching the search query.": [
            ""
        ],
        "Search within %(docstitle)s": "\u5728 %(docstitle)s \u4e2d\u641c\u5c0b",
        "Searching": "\u641c\u5c0b\u4e2d",
        "Searching for multiple words only shows matches that contain\n    all words.": "\u641c\u5c0b\u591a\u500b\u95dc\u9375\u5b57\u6642\uff0c\u53ea\u6703\u986f\u793a\u5305\u542b\u6240\u6709\u95dc\u9375\u5b57\u7684\u7d50\u679c\u3002",
        "Show Source": "\u986f\u793a\u539f\u59cb\u78bc",
        "Table of Contents": "\u76ee\u9304",
        "This Page": "\u672c\u9801",
        "Welcome! This is": "\u6b61\u8fce\uff01\u672c",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "\u60a8\u7684\u641c\u5c0b\u627e\u4e0d\u5230\u4efb\u4f55\u6eff\u8db3\u689d\u4ef6\u7684\u6587\u4ef6\u3002\u8acb\u78ba\u5b9a\u662f\u5426\u6240\u6709\u7684\u641c\u5c0b\u8a5e\u90fd\u6b63\u78ba\u5730\u62fc\u5beb\u4e14\u60a8\u5df2\u9078\u64c7\u8db3\u5920\u7684\u5206\u985e\u3002",
        "all functions, classes, terms": "\u6240\u6709\u51fd\u5f0f\u3001\u985e\u5225\u3001\u8853\u8a9e",
        "can be huge": "\u53ef\u80fd\u6703\u5f88\u5927",
        "last updated": "\u6700\u5f8c\u66f4\u65b0\u65bc",
        "lists all sections and subsections": "\u5217\u51fa\u6240\u6709\u7ae0\u7bc0\u8207\u5c0f\u7bc0",
        "next chapter": "\u4e0b\u4e00\u7ae0",
        "previous chapter": "\u4e0a\u4e00\u7ae0",
        "quick access to all modules": "\u8fc5\u901f\u627e\u5230\u6240\u6709\u6a21\u7d44",
        "search": "\u641c\u5c0b",
        "search this documentation": "\u641c\u5c0b\u9019\u4efd\u8aaa\u660e\u6587\u4ef6",
        "the documentation for": "\u8aaa\u660e\u6587\u4ef6\u4ecb\u7d39"
    },
    "plural_expr": "0"
});