# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/italian/stop.txt

from __future__ import annotations

ITALIAN_STOPWORDS = frozenset({
    'a',
    'abbia',
    'abbiamo',
    'abbiano',
    'abbiate',
    'ad',
    'agl',
    'agli',
    'ai',
    'al',
    'all',
    'alla',
    'alle',
    'allo',
    'anche',
    'avemmo',
    'avendo',
    'avesse',
    'avessero',
    'avessi',
    'avessimo',
    'aveste',
    'avesti',
    'avete',
    'aveva',
    'avevamo',
    'avevano',
    'avevate',
    'avevi',
    'avevo',
    'avrai',
    'avranno',
    'avrebbe',
    'avrebbero',
    'avrei',
    'avremmo',
    'avremo',
    'avreste',
    'avresti',
    'avrete',
    'avrà',
    'avrò',
    'avuta',
    'avute',
    'avuti',
    'avuto',
    'c',
    'che',
    'chi',
    'ci',
    'coi',
    'col',
    'come',
    'con',
    'contro',
    'cui',
    'da',
    'dagl',
    'dagli',
    'dai',
    'dal',
    'dall',
    'dalla',
    'dalle',
    'dallo',
    'degl',
    'degli',
    'dei',
    'del',
    'dell',
    'della',
    'delle',
    'dello',
    'di',
    'dov',
    'dove',
    'e',
    'ebbe',
    'ebbero',
    'ebbi',
    'ed',
    'era',
    'erano',
    'eravamo',
    'eravate',
    'eri',
    'ero',
    'essendo',
    'faccia',
    'facciamo',
    'facciano',
    'facciate',
    'faccio',
    'facemmo',
    'facendo',
    'facesse',
    'facessero',
    'facessi',
    'facessimo',
    'faceste',
    'facesti',
    'faceva',
    'facevamo',
    'facevano',
    'facevate',
    'facevi',
    'facevo',
    'fai',
    'fanno',
    'farai',
    'faranno',
    'farebbe',
    'farebbero',
    'farei',
    'faremmo',
    'faremo',
    'fareste',
    'faresti',
    'farete',
    'farà',
    'farò',
    'fece',
    'fecero',
    'feci',
    'fosse',
    'fossero',
    'fossi',
    'fossimo',
    'foste',
    'fosti',
    'fu',
    'fui',
    'fummo',
    'furono',
    'gli',
    'ha',
    'hai',
    'hanno',
    'ho',
    'i',
    'il',
    'in',
    'io',
    'l',
    'la',
    'le',
    'lei',
    'li',
    'lo',
    'loro',
    'lui',
    'ma',
    'mi',
    'mia',
    'mie',
    'miei',
    'mio',
    'ne',
    'negl',
    'negli',
    'nei',
    'nel',
    'nell',
    'nella',
    'nelle',
    'nello',
    'noi',
    'non',
    'nostra',
    'nostre',
    'nostri',
    'nostro',
    'o',
    'per',
    'perché',
    'più',
    'quale',
    'quanta',
    'quante',
    'quanti',
    'quanto',
    'quella',
    'quelle',
    'quelli',
    'quello',
    'questa',
    'queste',
    'questi',
    'questo',
    'sarai',
    'saranno',
    'sarebbe',
    'sarebbero',
    'sarei',
    'saremmo',
    'saremo',
    'sareste',
    'saresti',
    'sarete',
    'sarà',
    'sarò',
    'se',
    'sei',
    'si',
    'sia',
    'siamo',
    'siano',
    'siate',
    'siete',
    'sono',
    'sta',
    'stai',
    'stando',
    'stanno',
    'starai',
    'staranno',
    'starebbe',
    'starebbero',
    'starei',
    'staremmo',
    'staremo',
    'stareste',
    'staresti',
    'starete',
    'starà',
    'starò',
    'stava',
    'stavamo',
    'stavano',
    'stavate',
    'stavi',
    'stavo',
    'stemmo',
    'stesse',
    'stessero',
    'stessi',
    'stessimo',
    'steste',
    'stesti',
    'stette',
    'stettero',
    'stetti',
    'stia',
    'stiamo',
    'stiano',
    'stiate',
    'sto',
    'su',
    'sua',
    'sue',
    'sugl',
    'sugli',
    'sui',
    'sul',
    'sull',
    'sulla',
    'sulle',
    'sullo',
    'suo',
    'suoi',
    'ti',
    'tra',
    'tu',
    'tua',
    'tue',
    'tuo',
    'tuoi',
    'tutti',
    'tutto',
    'un',
    'una',
    'uno',
    'vi',
    'voi',
    'vostra',
    'vostre',
    'vostri',
    'vostro',
    'è',
})
