# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/dutch/stop.txt

from __future__ import annotations

DUTCH_STOPWORDS = frozenset({
    'aan',
    'al',
    'alles',
    'als',
    'altijd',
    'andere',
    'ben',
    'bij',
    'daar',
    'dan',
    'dat',
    'de',
    'der',
    'deze',
    'die',
    'dit',
    'doch',
    'doen',
    'door',
    'dus',
    'een',
    'eens',
    'en',
    'er',
    'ge',
    'geen',
    'geweest',
    'haar',
    'had',
    'heb',
    'hebben',
    'heeft',
    'hem',
    'het',
    'hier',
    'hij',
    'hoe',
    'hun',
    'iemand',
    'iets',
    'ik',
    'in',
    'is',
    'ja',
    'je',
    'kan',
    'kon',
    'kunnen',
    'maar',
    'me',
    'meer',
    'men',
    'met',
    'mij',
    'mijn',
    'moet',
    'na',
    'naar',
    'niet',
    'niets',
    'nog',
    'nu',
    'of',
    'om',
    'omdat',
    'onder',
    'ons',
    'ook',
    'op',
    'over',
    'reeds',
    'te',
    'tegen',
    'toch',
    'toen',
    'tot',
    'u',
    'uit',
    'uw',
    'van',
    'veel',
    'voor',
    'want',
    'waren',
    'was',
    'wat',
    'werd',
    'wezen',
    'wie',
    'wil',
    'worden',
    'wordt',
    'zal',
    'ze',
    'zelf',
    'zich',
    'zij',
    'zijn',
    'zo',
    'zonder',
    'zou',
})
