# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/danish/stop.txt

from __future__ import annotations

DANISH_STOPWORDS = frozenset({
    'ad',
    'af',
    'alle',
    'alt',
    'anden',
    'at',
    'blev',
    'blive',
    'bliver',
    'da',
    'de',
    'dem',
    'den',
    'denne',
    'der',
    'deres',
    'det',
    'dette',
    'dig',
    'din',
    'disse',
    'dog',
    'du',
    'efter',
    'eller',
    'en',
    'end',
    'er',
    'et',
    'for',
    'fra',
    'ham',
    'han',
    'hans',
    'har',
    'havde',
    'have',
    'hende',
    'hendes',
    'her',
    'hos',
    'hun',
    'hvad',
    'hvis',
    'hvor',
    'i',
    'ikke',
    'ind',
    'jeg',
    'jer',
    'jo',
    'kunne',
    'man',
    'mange',
    'med',
    'meget',
    'men',
    'mig',
    'min',
    'mine',
    'mit',
    'mod',
    'ned',
    'noget',
    'nogle',
    'nu',
    'når',
    'og',
    'også',
    'om',
    'op',
    'os',
    'over',
    'på',
    'selv',
    'sig',
    'sin',
    'sine',
    'sit',
    'skal',
    'skulle',
    'som',
    'sådan',
    'thi',
    'til',
    'ud',
    'under',
    'var',
    'vi',
    'vil',
    'ville',
    'vor',
    'være',
    'været',
})
