# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/spanish/stop.txt

from __future__ import annotations

SPANISH_STOPWORDS = frozenset({
    'a',
    'al',
    'algo',
    'algunas',
    'algunos',
    'ante',
    'antes',
    'como',
    'con',
    'contra',
    'cual',
    'cuando',
    'de',
    'del',
    'desde',
    'donde',
    'durante',
    'e',
    'el',
    'ella',
    'ellas',
    'ellos',
    'en',
    'entre',
    'era',
    'erais',
    'eran',
    'eras',
    'eres',
    'es',
    'esa',
    'esas',
    'ese',
    'eso',
    'esos',
    'esta',
    'estaba',
    'estabais',
    'estaban',
    'estabas',
    'estad',
    'estada',
    'estadas',
    'estado',
    'estados',
    'estamos',
    'estando',
    'estar',
    'estaremos',
    'estará',
    'estarán',
    'estarás',
    'estaré',
    'estaréis',
    'estaría',
    'estaríais',
    'estaríamos',
    'estarían',
    'estarías',
    'estas',
    'este',
    'estemos',
    'esto',
    'estos',
    'estoy',
    'estuve',
    'estuviera',
    'estuvierais',
    'estuvieran',
    'estuvieras',
    'estuvieron',
    'estuviese',
    'estuvieseis',
    'estuviesen',
    'estuvieses',
    'estuvimos',
    'estuviste',
    'estuvisteis',
    'estuviéramos',
    'estuviésemos',
    'estuvo',
    'está',
    'estábamos',
    'estáis',
    'están',
    'estás',
    'esté',
    'estéis',
    'estén',
    'estés',
    'fue',
    'fuera',
    'fuerais',
    'fueran',
    'fueras',
    'fueron',
    'fuese',
    'fueseis',
    'fuesen',
    'fueses',
    'fui',
    'fuimos',
    'fuiste',
    'fuisteis',
    'fuéramos',
    'fuésemos',
    'ha',
    'habida',
    'habidas',
    'habido',
    'habidos',
    'habiendo',
    'habremos',
    'habrá',
    'habrán',
    'habrás',
    'habré',
    'habréis',
    'habría',
    'habríais',
    'habríamos',
    'habrían',
    'habrías',
    'habéis',
    'había',
    'habíais',
    'habíamos',
    'habían',
    'habías',
    'han',
    'has',
    'hasta',
    'hay',
    'haya',
    'hayamos',
    'hayan',
    'hayas',
    'hayáis',
    'he',
    'hemos',
    'hube',
    'hubiera',
    'hubierais',
    'hubieran',
    'hubieras',
    'hubieron',
    'hubiese',
    'hubieseis',
    'hubiesen',
    'hubieses',
    'hubimos',
    'hubiste',
    'hubisteis',
    'hubiéramos',
    'hubiésemos',
    'hubo',
    'la',
    'las',
    'le',
    'les',
    'lo',
    'los',
    'me',
    'mi',
    'mis',
    'mucho',
    'muchos',
    'muy',
    'más',
    'mí',
    'mía',
    'mías',
    'mío',
    'míos',
    'nada',
    'ni',
    'no',
    'nos',
    'nosotras',
    'nosotros',
    'nuestra',
    'nuestras',
    'nuestro',
    'nuestros',
    'o',
    'os',
    'otra',
    'otras',
    'otro',
    'otros',
    'para',
    'pero',
    'poco',
    'por',
    'porque',
    'que',
    'quien',
    'quienes',
    'qué',
    'se',
    'sea',
    'seamos',
    'sean',
    'seas',
    'seremos',
    'será',
    'serán',
    'serás',
    'seré',
    'seréis',
    'sería',
    'seríais',
    'seríamos',
    'serían',
    'serías',
    'seáis',
    'sido',
    'siendo',
    'sin',
    'sobre',
    'sois',
    'somos',
    'son',
    'soy',
    'su',
    'sus',
    'suya',
    'suyas',
    'suyo',
    'suyos',
    'sí',
    'también',
    'tanto',
    'te',
    'tendremos',
    'tendrá',
    'tendrán',
    'tendrás',
    'tendré',
    'tendréis',
    'tendría',
    'tendríais',
    'tendríamos',
    'tendrían',
    'tendrías',
    'tened',
    'tenemos',
    'tenga',
    'tengamos',
    'tengan',
    'tengas',
    'tengo',
    'tengáis',
    'tenida',
    'tenidas',
    'tenido',
    'tenidos',
    'teniendo',
    'tenéis',
    'tenía',
    'teníais',
    'teníamos',
    'tenían',
    'tenías',
    'ti',
    'tiene',
    'tienen',
    'tienes',
    'todo',
    'todos',
    'tu',
    'tus',
    'tuve',
    'tuviera',
    'tuvierais',
    'tuvieran',
    'tuvieras',
    'tuvieron',
    'tuviese',
    'tuvieseis',
    'tuviesen',
    'tuvieses',
    'tuvimos',
    'tuviste',
    'tuvisteis',
    'tuviéramos',
    'tuviésemos',
    'tuvo',
    'tuya',
    'tuyas',
    'tuyo',
    'tuyos',
    'tú',
    'un',
    'una',
    'uno',
    'unos',
    'vosotras',
    'vosotros',
    'vuestra',
    'vuestras',
    'vuestro',
    'vuestros',
    'y',
    'ya',
    'yo',
    'él',
    'éramos',
})
