var SwedishStemmer=function(){var a=new BaseStemmer,i=[["fab",-1,-1],["h",-1,-1],["pak",-1,-1],["rak",-1,-1],["stak",-1,-1],["kom",-1,-1],["iet",-1,-1],["cit",-1,-1],["dit",-1,-1],["alit",-1,-1],["ilit",-1,-1],["mit",-1,-1],["nit",-1,-1],["pit",-1,-1],["rit",-1,-1],["sit",-1,-1],["tit",-1,-1],["uit",-1,-1],["ivit",-1,-1],["kvit",-1,-1],["xit",-1,-1]],e=[["a",-1,1],["arna",0,1],["erna",0,1],["heterna",2,1],["orna",0,1],["ad",-1,1],["e",-1,1],["ade",6,1],["ande",6,1],["arne",6,1],["are",6,1],["aste",6,1],["en",-1,1],["anden",12,1],["aren",12,1],["heten",12,1],["ern",-1,1],["ar",-1,1],["er",-1,1],["heter",18,1],["or",-1,1],["s",-1,2],["as",21,1],["arnas",22,1],["ernas",22,1],["ornas",22,1],["es",21,1],["ades",26,1],["andes",26,1],["ens",21,1],["arens",29,1],["hetens",29,1],["erns",21,1],["at",-1,1],["et",-1,3],["andet",34,1],["het",34,1],["ast",-1,1]],s=[["dd",-1,-1],["gd",-1,-1],["nn",-1,-1],["dt",-1,-1],["gt",-1,-1],["kt",-1,-1],["tt",-1,-1]],c=[["ig",-1,1],["lig",0,1],["els",-1,1],["fullt",-1,3],["öst",-1,2]],u=[17,65,16,1,0,0,0,0,0,0,0,0,0,0,0,0,24,0,32],o=[119,127,149],n=[173,58],l=0,m=0;function _(){var r=a.limit-a.cursor;if(a.out_grouping_b(u,97,246)&&(a.in_grouping_b(u,97,246)&&a.cursor>a.limit_backward))return a.cursor=a.limit-r,r=a.limit-a.cursor,0==a.find_among_b(i)?(a.cursor=a.limit-r,1):void 0}this.stem=function(){var r=a.cursor,i=(m=a.limit,t=a.cursor,(i=a.cursor+3)>a.limit||(a.cursor=i,l=a.cursor,a.cursor=t,a.go_out_grouping(u,97,246)&&(a.cursor++,a.go_in_grouping(u,97,246))&&(a.cursor++,m=a.cursor,l<=m||(m=l))),a.cursor=r,a.limit_backward=a.cursor,a.cursor=a.limit,a.limit-a.cursor),t=((()=>{var r;if(!(a.cursor<m)){var i=a.limit_backward;if(a.limit_backward=m,a.ket=a.cursor,0==(r=a.find_among_b(e)))a.limit_backward=i;else switch(a.bra=a.cursor,a.limit_backward=i,r){case 1:if(a.slice_del())break;return;case 2:var t=a.limit-a.cursor;if(a.eq_s_b("et")&&_())a.bra=a.cursor;else if(a.cursor=a.limit-t,!a.in_grouping_b(o,98,121))return;if(a.slice_del())break;return;case 3:if(!_())return;if(a.slice_del())break}}})(),a.cursor=a.limit-i,a.limit-a.cursor),i=(a.cursor<m||(r=a.limit_backward,a.limit_backward=m,i=a.limit-a.cursor,(0==a.find_among_b(s)||(a.cursor=a.limit-i,a.ket=a.cursor,a.cursor<=a.limit_backward)||(a.cursor--,a.bra=a.cursor,a.slice_del()))&&(a.limit_backward=r)),a.cursor=a.limit-t,a.limit-a.cursor);return(()=>{var r;if(!(a.cursor<m)){var i=a.limit_backward;if(a.limit_backward=m,a.ket=a.cursor,0==(r=a.find_among_b(c)))a.limit_backward=i;else switch(a.bra=a.cursor,a.limit_backward=i,r){case 1:if(a.slice_del())break;return;case 2:if(!a.in_grouping_b(n,105,118))return;if(a.slice_from("ös"))break;return;case 3:if(a.slice_from("full"))break}}})(),a.cursor=a.limit-i,a.cursor=a.limit_backward,!0},this.stemWord=function(r){return a.setCurrent(r),this.stem(),a.getCurrent()}};