Documentation.addTranslations({
    "locale": "ca@valencia",
    "messages": {
        "%(filename)s &#8212; %(docstitle)s": "%(filename)s &#8212; %(docstitle)s",
        "&#169; %(copyright_prefix)s %(copyright)s.": "&#169; %(copyright_prefix)s %(copyright)s.",
        ", in ": ", a ",
        "About these documents": "Quant a estos documents",
        "Automatically generated list of changes in version %(version)s": "Llista de canvis de la versi\u00f3 %(version)s generada autom\u00e0ticament",
        "C API changes": "Canvis a l'API de C",
        "Changes in Version %(version)s &#8212; %(docstitle)s": "Canvis en la versi\u00f3 %(version)s &#8212; %(docstitle)s",
        "Collapse sidebar": "Redu\u00efx la barra lateral",
        "Complete Table of Contents": "Taula de contingut completa",
        "Contents": "Contingut",
        "Copyright": "Copyright",
        "Created using <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "Creada mitjan\u00e7ant <a href=\"https://www.sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.",
        "Expand sidebar": "Expandix la barra lateral",
        "Full index on one page": "\u00cdndex complet en una p\u00e0gina",
        "General Index": "\u00cdndex general",
        "Global Module Index": "\u00cdndex global de m\u00f2duls",
        "Go": "Ves fins a",
        "Hide Search Matches": "Oculta els resultats de busca",
        "Index": "\u00cdndex",
        "Index &#x2013; %(key)s": "\u00cdndex &#x2013; %(key)s",
        "Index pages by letter": "P\u00e0gines d'\u00edndex per lletra",
        "Indices and tables:": "\u00cdndexs i taules:",
        "Last updated on %(last_updated)s.": "\u00daltima actualitzaci\u00f3 el %(last_updated)s.",
        "Library changes": "Canvis a la biblioteca",
        "Navigation": "Navegueu",
        "Next topic": "Tema seg\u00fcent",
        "Other changes": "Altres canvis",
        "Overview": "Vista general",
        "Please activate JavaScript to enable the search\n    functionality.": "Activa JavaScript per a utilitzar la funcionalitat\n    de busca.",
        "Preparing search...": "S'est\u00e0 preparant la busca\u2026",
        "Previous topic": "Tema anterior",
        "Quick search": "Busca r\u00e0pida",
        "Search": "Busca",
        "Search Page": "P\u00e0gina de busca",
        "Search Results": "Resultats de la busca",
        "Search finished, found one page matching the search query.": [
            "La busca ha acabat, s'ha trobat una p\u00e0gina que coincidix amb la consulta de busca.",
            "La busca ha acabat, s'han trobat ${resultCount} p\u00e0gines que coincidixen amb la consulta de busca."
        ],
        "Search within %(docstitle)s": "Busca dins de %(docstitle)s",
        "Searching": "S'est\u00e0 buscant",
        "Searching for multiple words only shows matches that contain\n    all words.": "Buscar m\u00faltiples paraules nom\u00e9s mostrar\u00e0 les coincid\u00e8ncies\n    que continguen totes les paraules.",
        "Show Source": "Mostra el codi font",
        "Table of Contents": "Taula de continguts",
        "This Page": "Esta p\u00e0gina",
        "Welcome! This is": "Us donem la benvinguda! A\u00e7\u00f2 \u00e9s",
        "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "La vostra busca no ha coincidit amb cap document. Assegureu-vos que s'escriuen correctament totes les paraules i que heu seleccionat prou categories.",
        "all functions, classes, terms": "totes les funcions, classes, termes",
        "can be huge": "pot ser gegant",
        "last updated": "\u00faltima actualitzaci\u00f3",
        "lists all sections and subsections": "llista totes les seccions i subseccions",
        "next chapter": "cap\u00edtol seg\u00fcent",
        "previous chapter": "cap\u00edtol anterior",
        "quick access to all modules": "acc\u00e9s r\u00e0pid a tots els m\u00f2duls",
        "search": "busca",
        "search this documentation": "busca esta documentaci\u00f3",
        "the documentation for": "la documentaci\u00f3 per a"
    },
    "plural_expr": "(n != 1)"
});