var NorwegianStemmer=function(){var t=new BaseStemmer,a=[["",-1,1],["ind",0,-1],["kk",0,-1],["nk",0,-1],["amm",0,-1],["omm",0,-1],["kap",0,-1],["skap",6,1],["pp",0,-1],["lt",0,-1],["ast",0,-1],["øst",0,-1],["v",0,-1],["hav",12,1],["giv",12,1]],c=[["a",-1,1],["e",-1,1],["ede",1,1],["ande",1,1],["ende",1,1],["ane",1,1],["ene",1,1],["hetene",6,1],["erte",1,4],["en",-1,1],["heten",9,1],["ar",-1,1],["er",-1,1],["heter",12,1],["s",-1,3],["as",14,1],["es",14,1],["edes",16,1],["endes",16,1],["enes",16,1],["hetenes",19,1],["ens",14,1],["hetens",21,1],["ers",14,2],["ets",14,1],["et",-1,1],["het",25,1],["ert",-1,4],["ast",-1,1]],s=[["dt",-1,-1],["vt",-1,-1]],o=[["leg",-1,1],["eleg",0,1],["ig",-1,1],["eig",2,1],["lig",2,1],["elig",4,1],["els",-1,1],["lov",-1,1],["elov",7,1],["slov",7,1],["hetslov",9,1]],u=[17,65,16,1,0,0,0,0,0,0,0,0,0,0,0,0,48,2,142],l=[119,125,148,1],m=0,n=0;this.stem=function(){var r=t.cursor,i=(n=t.limit,e=t.cursor,(i=t.cursor+3)>t.limit||(t.cursor=i,m=t.cursor,t.cursor=e,t.go_out_grouping(u,97,248)&&(t.cursor++,t.go_in_grouping(u,97,248))&&(t.cursor++,n=t.cursor,m<=n||(n=m))),t.cursor=r,t.limit_backward=t.cursor,t.cursor=t.limit,t.limit-t.cursor),e=((()=>{var r;if(!(t.cursor<n)){var i=t.limit_backward;if(t.limit_backward=n,t.ket=t.cursor,0==(r=t.find_among_b(c)))t.limit_backward=i;else switch(t.bra=t.cursor,t.limit_backward=i,r){case 1:if(t.slice_del())break;return;case 2:switch(t.find_among_b(a)){case 1:if(t.slice_del())break;return}break;case 3:r:{var e=t.limit-t.cursor;if(!t.in_grouping_b(l,98,122)){if(t.cursor=t.limit-e,t.eq_s_b("r")){var s=t.limit-t.cursor;if(!t.eq_s_b("e")){t.cursor=t.limit-s;break r}}if(t.cursor=t.limit-e,!t.eq_s_b("k"))return;if(!t.out_grouping_b(u,97,248))return}}if(t.slice_del())break;return;case 4:if(t.slice_from("er"))break}}})(),t.cursor=t.limit-i,t.limit-t.cursor),i=(r=t.limit-t.cursor,t.cursor<n||(i=t.limit_backward,t.limit_backward=n,t.ket=t.cursor,0==t.find_among_b(s)?t.limit_backward=i:(t.bra=t.cursor,t.limit_backward=i,t.cursor=t.limit-r,t.cursor<=t.limit_backward||(t.cursor--,t.bra=t.cursor,t.slice_del()))),t.cursor=t.limit-e,t.limit-t.cursor);return t.cursor<n||(r=t.limit_backward,t.limit_backward=n,t.ket=t.cursor,0==t.find_among_b(o)?t.limit_backward=r:(t.bra=t.cursor,t.limit_backward=r,t.slice_del())),t.cursor=t.limit-i,t.cursor=t.limit_backward,!0},this.stemWord=function(r){return t.setCurrent(r),this.stem(),t.getCurrent()}};