# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/finnish/stop.txt

from __future__ import annotations

FINNISH_STOPWORDS = frozenset({
    'ei',
    'eivät',
    'emme',
    'en',
    'et',
    'ette',
    'että',
    'he',
    'heidän',
    'heidät',
    'heihin',
    'heille',
    'heillä',
    'heiltä',
    'heissä',
    'heistä',
    'heitä',
    'hän',
    'häneen',
    'hänelle',
    'hänellä',
    'häneltä',
    'hänen',
    'hänessä',
    'hänestä',
    'hänet',
    'häntä',
    'itse',
    'ja',
    'johon',
    'joiden',
    'joihin',
    'joiksi',
    'joilla',
    'joille',
    'joilta',
    'joina',
    'joissa',
    'joista',
    'joita',
    'joka',
    'joksi',
    'jolla',
    'jolle',
    'jolta',
    'jona',
    'jonka',
    'jos',
    'jossa',
    'josta',
    'jota',
    'jotka',
    'kanssa',
    'keiden',
    'keihin',
    'keiksi',
    'keille',
    'keillä',
    'keiltä',
    'keinä',
    'keissä',
    'keistä',
    'keitä',
    'keneen',
    'keneksi',
    'kenelle',
    'kenellä',
    'keneltä',
    'kenen',
    'kenenä',
    'kenessä',
    'kenestä',
    'kenet',
    'ketkä',
    'ketä',
    'koska',
    'kuin',
    'kuka',
    'kun',
    'me',
    'meidän',
    'meidät',
    'meihin',
    'meille',
    'meillä',
    'meiltä',
    'meissä',
    'meistä',
    'meitä',
    'mihin',
    'miksi',
    'mikä',
    'mille',
    'millä',
    'miltä',
    'minkä',
    'minua',
    'minulla',
    'minulle',
    'minulta',
    'minun',
    'minussa',
    'minusta',
    'minut',
    'minuun',
    'minä',
    'missä',
    'mistä',
    'mitkä',
    'mitä',
    'mukaan',
    'mutta',
    'ne',
    'niiden',
    'niihin',
    'niiksi',
    'niille',
    'niillä',
    'niiltä',
    'niin',
    'niinä',
    'niissä',
    'niistä',
    'niitä',
    'noiden',
    'noihin',
    'noiksi',
    'noilla',
    'noille',
    'noilta',
    'noin',
    'noina',
    'noissa',
    'noista',
    'noita',
    'nuo',
    'nyt',
    'näiden',
    'näihin',
    'näiksi',
    'näille',
    'näillä',
    'näiltä',
    'näinä',
    'näissä',
    'näistä',
    'näitä',
    'nämä',
    'ole',
    'olemme',
    'olen',
    'olet',
    'olette',
    'oli',
    'olimme',
    'olin',
    'olisi',
    'olisimme',
    'olisin',
    'olisit',
    'olisitte',
    'olisivat',
    'olit',
    'olitte',
    'olivat',
    'olla',
    'olleet',
    'ollut',
    'on',
    'ovat',
    'poikki',
    'se',
    'sekä',
    'sen',
    'siihen',
    'siinä',
    'siitä',
    'siksi',
    'sille',
    'sillä',
    'siltä',
    'sinua',
    'sinulla',
    'sinulle',
    'sinulta',
    'sinun',
    'sinussa',
    'sinusta',
    'sinut',
    'sinuun',
    'sinä',
    'sitä',
    'tai',
    'te',
    'teidän',
    'teidät',
    'teihin',
    'teille',
    'teillä',
    'teiltä',
    'teissä',
    'teistä',
    'teitä',
    'tuo',
    'tuohon',
    'tuoksi',
    'tuolla',
    'tuolle',
    'tuolta',
    'tuon',
    'tuona',
    'tuossa',
    'tuosta',
    'tuota',
    'tähän',
    'täksi',
    'tälle',
    'tällä',
    'tältä',
    'tämä',
    'tämän',
    'tänä',
    'tässä',
    'tästä',
    'tätä',
    'vaan',
    'vai',
    'vaikka',
    'yli',
})
