# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/hungarian/stop.txt

from __future__ import annotations

HUNGARIAN_STOPWORDS = frozenset({
    'a',
    'abban',
    'ahhoz',
    'ahogy',
    'ahol',
    'aki',
    'akik',
    'akkor',
    'alatt',
    'amely',
    'amelyek',
    'amelyekben',
    'amelyeket',
    'amelyet',
    'amelynek',
    'ami',
    'amikor',
    'amit',
    'amolyan',
    'amíg',
    'annak',
    'arra',
    'arról',
    'az',
    'azok',
    'azon',
    'azonban',
    'azt',
    'aztán',
    'azután',
    'azzal',
    'azért',
    'be',
    'belül',
    'benne',
    'bár',
    'cikk',
    'cikkek',
    'cikkeket',
    'csak',
    'de',
    'e',
    'ebben',
    'eddig',
    'egy',
    'egyes',
    'egyetlen',
    'egyik',
    'egyre',
    'egyéb',
    'egész',
    'ehhez',
    'ekkor',
    'el',
    'ellen',
    'elsõ',
    'elég',
    'elõ',
    'elõször',
    'elõtt',
    'emilyen',
    'ennek',
    'erre',
    'ez',
    'ezek',
    'ezen',
    'ezt',
    'ezzel',
    'ezért',
    'fel',
    'felé',
    'hanem',
    'hiszen',
    'hogy',
    'hogyan',
    'igen',
    'ill',
    'ill.',
    'illetve',
    'ilyen',
    'ilyenkor',
    'ismét',
    'ison',
    'itt',
    'jobban',
    'jó',
    'jól',
    'kell',
    'kellett',
    'keressünk',
    'keresztül',
    'ki',
    'kívül',
    'között',
    'közül',
    'legalább',
    'legyen',
    'lehet',
    'lehetett',
    'lenne',
    'lenni',
    'lesz',
    'lett',
    'maga',
    'magát',
    'majd',
    'meg',
    'mellett',
    'mely',
    'melyek',
    'mert',
    'mi',
    'mikor',
    'milyen',
    'minden',
    'mindenki',
    'mindent',
    'mindig',
    'mint',
    'mintha',
    'mit',
    'mivel',
    'miért',
    'most',
    'már',
    'más',
    'másik',
    'még',
    'míg',
    'nagy',
    'nagyobb',
    'nagyon',
    'ne',
    'nekem',
    'neki',
    'nem',
    'nincs',
    'néha',
    'néhány',
    'nélkül',
    'olyan',
    'ott',
    'pedig',
    'persze',
    'rá',
    's',
    'saját',
    'sem',
    'semmi',
    'sok',
    'sokat',
    'sokkal',
    'szemben',
    'szerint',
    'szinte',
    'számára',
    'talán',
    'tehát',
    'teljes',
    'tovább',
    'továbbá',
    'több',
    'ugyanis',
    'utolsó',
    'után',
    'utána',
    'vagy',
    'vagyis',
    'vagyok',
    'valaki',
    'valami',
    'valamint',
    'való',
    'van',
    'vannak',
    'vele',
    'vissza',
    'viszont',
    'volna',
    'volt',
    'voltak',
    'voltam',
    'voltunk',
    'által',
    'általában',
    'át',
    'én',
    'éppen',
    'és',
    'így',
    'õ',
    'õk',
    'õket',
    'össze',
    'úgy',
    'új',
    'újabb',
    'újra',
})
